<?php
/**
 * Magento Module developed by NoStress Commerce
 *
 * NOTICE OF LICENSE
 *
 * This program is licensed under the Koongo software licence (by NoStress Commerce). 
 * With the purchase, download of the software or the installation of the software 
 * in your application you accept the licence agreement. The allowed usage is outlined in the
 * Koongo software licence which can be found under https://docs.koongo.com/display/koongo/License+Conditions
 *
 * Any modification or distribution is strictly forbidden. The license
 * grants you the installation in one application. For multiuse you will need
 * to purchase further licences at https://store.koongo.com/.
 *
 * See the Koongo software licence agreement for more details.
 * @copyright Copyright (c) 2017 NoStress Commerce (http://www.nostresscommerce.cz, http://www.koongo.com/)
 *
 */

/**
 * Model for Koongo API stock repository
 *
 * @category Nostress
 * @package Nostress_Koongo
 *
 */

namespace Nostress\Koongo\Model\Cache;

use Nostress\Koongo\Api\Cache\StockRepositoryInterface;
use Magento\Framework\Exception\InputException;
use Magento\Framework\Exception\LocalizedException;
use Magento\Framework\Exception\NoSuchEntityException;
use Magento\Framework\Exception\StateException;
use Magento\Framework\Exception\ValidatorException;
use Magento\Framework\Exception\CouldNotSaveException;
use Nostress\Koongo\Model\ResourceModel\Cache\Stock\Collection;
use Magento\Framework\Api\SortOrder;

class StockRepository implements StockRepositoryInterface
{       
    /**
     * Stock factory
     *
	 * @var \Nostress\Koongo\Model\Cache\StockFactory 
	 */
    protected $_stockFactory;

    /**
     * Stock resource model
     *
     * @var \Nostress\Koongo\Model\ResourceModel\Cache\Stock
     */
    protected $_resourceModel;
    
    /**
     * @var \Nostress\Koongo\Api\Cache\Data\StockSearchResultsInterfaceFactory
     */
    protected $_searchResultsFactory;

    /**
     *
     * @var \Nostress\Koongo\Helper
     */
    public $helper;

    /** 
     * @param \Nostress\Koongo\Model\Cache\StockFactory $stockFactory
     * @param \Nostress\Koongo\Model\ResourceModel\Cache\Stock $resourceModel,
	 */
	public function __construct(               
        \Nostress\Koongo\Model\Cache\StockFactory $stockFactory,
        \Nostress\Koongo\Model\ResourceModel\Cache\Stock $resourceModel,
        \Nostress\Koongo\Api\Cache\Data\StockSearchResultsInterfaceFactory $searchResultsFactory,
        \Nostress\Koongo\Helper\Data $helper
        )
    {	            
        $this->_stockFactory = $stockFactory;
        $this->_resourceModel = $resourceModel;        
        $this->_searchResultsFactory = $searchResultsFactory;
        $this->helper = $helper;
    }    

    /**
     * Get stock record list
     *
     * @param \Magento\Framework\Api\SearchCriteriaInterface $searchCriteria
     * @return \Nostress\Koongo\Api\Cache\Data\StockSearchResultsInterface
     */
    public function getList(\Magento\Framework\Api\SearchCriteriaInterface $searchCriteria)
    {
        /** @var \Nostress\Koongo\Model\ResourceModel\Cache\Stock\Collection $collection */
        $collection = $this->_stockFactory->create()->getCollection();

        //Add filters from root filter group to the collection
        foreach ($searchCriteria->getFilterGroups() as $group) {
            $this->addFilterGroupToCollection($group, $collection);
        }

        /** @var SortOrder $sortOrder */
        foreach ((array)$searchCriteria->getSortOrders() as $sortOrder) {
            $field = $sortOrder->getField();
            $collection->addOrder(
                $field,
                ($sortOrder->getDirection() == SortOrder::SORT_ASC) ? 'ASC' : 'DESC'
            );
        }
        $collection->setCurPage($searchCriteria->getCurrentPage());
        $collection->setPageSize($searchCriteria->getPageSize());        
        $collection->load();

        $searchResult = $this->_searchResultsFactory->create();
        $searchResult->setSearchCriteria($searchCriteria);
        $searchResult->setItems($collection->getItems());
        $searchResult->setTotalCount($collection->getSize());
        return $searchResult;
    }

    /**
     * Helper function that adds a FilterGroup to the collection.
     *
     * @param \Magento\Framework\Api\Search\FilterGroup $filterGroup
     * @param Collection $collection
     * @return void
     */
    protected function addFilterGroupToCollection(
        \Magento\Framework\Api\Search\FilterGroup $filterGroup,
        Collection $collection
    ) {        
        foreach ($filterGroup->getFilters() as $filter) {
            $conditionType = $filter->getConditionType() ? $filter->getConditionType() : 'eq';
            $collection->addFieldToFilter($filter->getField(),[$conditionType => $filter->getValue()]);            
        }        
    }

    /**
     * Reload stock data for given store
     *
     * @param int $storeId
     * @return string Json config.
     */
    public function getReload($storeId)
    {
        $status = "success"; 
        $message = "";
        try {
            $this->_stockFactory->create()->reload($storeId);

        } catch (\Exception $e) 
        {
            $status = "failure";
            $message = $e->getMessage();
            $this->helper->log("Koongo stock cache reload failure for store_id ".$storeId);
        }                    

        $result = ['status' => $status, 'message' => $message];                
		return array($result);
    }
}